/*
 *
 *                  Copyright 1991,1992 Premia Corp.
 *                        All rights reserved
 *
 *    This source code is provided to the purchaser of the product
 *    for the purpose of allowing the purchaser to customize and/or
 *    extend the functionality of the product.  The source code   
 *    remains the property of Premia Corp. with all rights reserved.
 *    Any modifications to the source code are considered derivative
 *    works and all rights thereto are reserved to Premia Corp. except
 *    that the purchaser may use the modified product in the same
 *    permitted manner as the unmodified product.  It is expressly
 *    prohibited to distribute in any manner any part of the original
 *    source code, whether in source or object form, without the
 *    express written permission from Premia Corp.
 *
 *    Owner:
 *        Premia Corporation
 *        1075 NW Murray Blvd., Suite 268
 *        Portland OR 97229
 *
 */
/* All code was modified from code provided by Premia Corp. */
/* 9-Oct-1993  - James Davis */


#include <windows.h>
#include <string.h>
#include <ctype.h>
#include "exports.h"
#include "cwstart.h"

LIBMAIN



#define MAX_BUF_LEN    100

static LPVOID _lsp_hash_names        = NIL;  /* pointer to the hash table containing */
                                             /* all 'lsp' keyword names. */

static LPVOID _lsp_hash_templates    = NIL;  /* pointer to the hash table containing */
                                             /* an abbreviated name and its template */

/* Define a structure to hold an abbreviation and the template to be
 * expanded when the abbreviation is entered into the buffer.
 * This structure will be the element in the hash table _lsp_hash_templates.
 */
typedef struct _LanguageTemplate {         
   LPSTR abbrev;
   LPSTR template;
} LanguageTemplate;


/* This string contains all keyword names which will be indented an extra level.
 */
LPSTR LSP_INDENT_STR = "PROGN\377FOR\377IF\377SETQ\377WHILE\377COND\377";

/*
 * _lsp_templates[] contains the keyword abbreviation and the template for the
 * abbreviation.  To add a new name, insert it before the NIL entry or
 * call _lsp_assign_template().
 *
 * The special characters within the template are:
 *
 *     \n - new line
 *     &  - cursor position after template insertion
 *     @  - backspace
 *     \c - insert 'c' literally
 */
LanguageTemplate _lsp_templates[] = {
   { "(if",      "(if & \n(progn\n);progn\n);if" },
   { "(defun",   "(defun &(/)\n);defun" },
   { "(def",     "(defun &(/)\n);defun" },
   { "(progn",   "\t(progn &\t\n);progn" },
   { "(setq",    "(setq &\n);setq" },
   { "(while",   "(while &\n);while" },
   { "(",        "(&\n);" },
   { "(prompt",   "(prompt \"\n&\");prompt" },
   { "(getv",    "(getvar \"&\");getvar" },
   { "(setv",    "(setvar \"&\");getvar" },
   { "(cond",    "(cond \n ((&)\n\n)\n);cond" },
   { "(getvar",    "(getvar \"&\");getvar" },
   { "(setvar",    "(setvar \"&\");getvar" },
   { "(foreach", "(foreach &\n);foreach" },
   { NIL,         NIL },
};

/* Define the list of 'lsp' keyword names to appear in
 * the keyword color if language coloring is enabled.
 * This list contains the name of some functions that I
 * have written and that I wanted to color highliht.
 * Make this list as short as you need to improve speed
 * (although I'm not too sure that will be nesc.)
 */
LPSTR _lsp_names[] = {

   "ABORT",       "ABS",         "ACADVER",        "ALERT",
   "ALLOC",       "ALLOW",       "ALLTRIM",        "AND",
   "ANGLE",       "ANGTOF",      "ANGTOS",         "APPEND",
   "APPLY",       "ASCII",       "ASSOC",          "ATAN",
   "ATOF",        "ATOHEX",      "ATOI",           "ATOMS-FAMILY",
   "ATOM",        "AT",          "BITS",           "BLANKSYM",
   "BOOLE",       "BOUNDP",      "BOX",            "BREAK",
   "CAAAAR",      "CAAADR",      "CAAAR",          "CAADAR",
   "CAADDR",      "CAADR",       "CAAR",           "CADAAR",
   "CADADR",      "CADAR",       "CADDAR",         "CADDDR",
   "CADDR",       "CADR",        "CAR",            "CDAAAR",
   "CDAADR",      "CDAAR",       "CDADAR",         "CDADDR",
   "CDADR",       "CDAR",        "CDATA",          "CDATE",
   "CDDAAR",      "CDDADR",      "CDDAR",          "CDDDAR",
   "CDDDDR",      "CDDDR",       "CDDR",           "CDR",
   "CHR",         "CLOSE",       "COMMAND",        "COMMENT",
   "COND",        "CONF",        "CONS",           "COS",
   "CURENT",      "DATA",        "DATE",           "DAT",
   "DAY",         "DCHK",        "DEFUN",          "DELTA",
   "DEMO",        "DIATEMP",     "DIFFERENCE",     "DIRECT",
   "DIR",         "DISTANCE",    "DIST",           "DKEY",
   "DLAST",       "DLINE",       "DMSG",           "DNAME",
   "DONE",        "DOT",         "DPT",            "DXF",
   "EDITBOX",     "ENTDEL",      "ENTGETX",        "ENTGET",
   "ENTLAST"      "ENTMOD",      "ENTNEXT",        "ENTSEL",
   "ENTUPD",      "EQUAL",       "ERCMD",          "ERMSG",
   "ERRORA",      "ERRORT",      "EVALHOST",       "EVAL",
   "EXIT",        "EXPAND",      "EXPT",           "EXP",
   "FILE",        "FINDFILE",    "FLAG",           "FLOAT",
   "FLOW",        "FNAME",       "FNT",            "FOREACH",
   "FOUND",       "FTRAP",       "FUNCTION",       "FUNCT",
   "GCD",         "GC",          "GETALL",         "GETANGLE",
   "GETATT",      "GETBOX",      "GETCOD",         "GETCON",
   "GETCORNER",   "GETDIST",     "GETENV",         "GETEPT",
   "GETHOST",     "GETINT",      "GETKWORD",       "GETORIENT",
   "GETPOINT",    "GETREAL",     "GETSTRING",      "GETTAG",
   "GETVAR",      "GOOD",        "GRAPHSCR",       "GRCLEAR",
   "GRDRAW",      "GREATERP",    "GRREAD",         "GRTEXT",
   "HANDENT",     "HAND",        "HDRENT",         "HEXTOA",
   "HIGHLI",      "HOSTID",      "HOUR",           "ID",
   "IF",          "INITGET",     "INLOCK",         "INTERS",
   "INT",         "ITOA",        "JST",            "KEY",
   "KMAP",        "KW",          "LAMBDA",         "LAMBDA#01",
   "LAMBDA#02",   "LAMBDA#03",   "LAMBDA#04",      "LAMBDA#05",
   "LAMBDA#06",   "LAMBDA#07",   "LAMBDA#08",      "LAMBDA#09",
   "LAMBDA#10",   "LAST",        "LAYER",          "LCREATE",
   "LDEF",        "LEADER",      "LEADER",         "LENCHK",
   "LENGTH",      "LEN",         "LESSP",          "LF",
   "LINE",        "LISTP",       "LIST",           "LNAME",
   "LOAD",        "LOGAND",      "LOGIOR",         "LOG",
   "MAPCAR",      "MATCH",       "MAX",            "MEMBER",
   "MEM",         "MENUCMD",     "MINUSP",         "MINUS",
   "MIN",         "MLAYER",      "MODE",           "MONTH",
   "MSG",         "NAME",        "NDX",            "NENTSELP",
   "NENTSEL",     "NEWPRI",      "NEW",            "NILL",
   "NORM",        "NOT",         "NTH",            "NULL",
   "NUMBERP",     "NUM",         "OBJ",            "OFST",
   "OPEN",        "OSMODE",      "OSNAP",          "PAUSE",
   "PERPAN",      "PERPI",       "PERROR",         "PICKSET",
   "PLUS",        "POLAR",       "POS",            "PREV",
   "PRIN1",       "PRINC",       "PRINT",          "PROGN",
   "PROMPT",      "QUIT",        "QUOTE",          "QUOTIENT",
   "RCODE",       "READ-CHAR",   "READ-LINE",      "READ",
   "REAL",        "REDRAW",      "REGAPP",         "REGEN",
   "REMAINDER",   "REM",         "REPEAT",         "REPLY",
   "RESP",        "REVERSE",     "REV",            "RTOS",
   "SCAN",        "SEARCH",      "SETDFLT",        "SETLAYER",
   "SETQ",        "SETSCALE",    "SETUNIT",        "SETVAR",
   "SET",         "SHELL",       "SINVAR",         "SIN",
   "SNAPMO",      "SNAPUN",      "SNAP",           "SORTENTS",
   "SQRT",        "SRCLIN",      "SSADD",          "SSDEL",
   "SSGET",       "SSLENGTH",    "SSMEMB",         "SSNAME",
   "SSTR",        "STAT",        "STRCASE",        "STRCAT",
   "STRLEN",      "STRSCN",      "STRVAL",         "STR",
   "SUBSTR",      "SUBST",       "SUM",            "SYMBOL",
   "TBLNEXT",     "TBLSEARCH",   "TERPRI",         "TEXTSCR",
   "TIMES",       "TKEY",        "TLYR",           "TRACE",
   "TRANS",       "TRAP",        "TRIM",           "TYPE",
   "UBFLAG",      "UNDO",        "UNTRACE",        "UNTRAP",
   "VAL",         "VAR",         "VER",            "VMON",
   "VPORTS",      "WHILE",       "WRITE-CHAR",     "WRITE-LINE",
   "XLOAD",       "YEAR",        "ZEROP",        NIL              
};

static long _lsp_in_comment( long offset );

void DLL _lsp_init( void );
BOOL DLL _lsp_expand( void );
long DLL _lsp_language_colors_update( long firstline, long lastline );
void DLL _lsp_language_colors( long numlines );
void DLL _lsp_assign_template( LPSTR abbrev, LPSTR template );
void DLL _lsp_add_keyword( LPSTR keyword );
BOOL DLL _lsp_indent( void );


void DLL 
_init()
{
   _lsp_init();
}



/*
 ** _lsp_init()
 *
 *  Initialize this module by exported any required functions and
 *  the hash tables needed to support indenting, template expandion
 *  and keyword coloring.
 */
void DLL
_lsp_init( void )
{
   static BOOL firstTime = TRUE;
   int    i;

   if (firstTime)
   {
      LibExport( "BOOL _lsp_expand");
      LibExport( "_lsp_assign_template LPSTR LPSTR");
      LibExport( "_lsp_add_keyword LPSTR");
      LibExport( "BOOL _lsp_indent");
      LibExport( "long _lsp_language_colors_update long long");
      LibExport( "_lsp_language_colors long");

      firstTime = FALSE;

      /* Create a hash table to hold all of the 'lsp' keyword names.
       */
      _lsp_hash_names = HashCreateTable( 701 /* a prime # */, sizeof( LPSTR ), HASH_IGN_CASE );
      i = 0;

      while ( _lsp_names[ i ] != NIL )
      {
         LPSTR XFAR *lpf = (LPSTR XFAR *)HashGetEntry( _lsp_hash_names, 
               _lsp_names[i], strlen( _lsp_names[i] ) );
         if (lpf)
         {
            *lpf = StrNew( _lsp_names[i] );
            i++;
         }
      }

      /* Create a hash table to hold all of the 'lsp' template names.
       */
      _lsp_hash_templates = HashCreateTable( 37 /* a prime # */, sizeof( LPSTR ), HASH_IGN_CASE );
      i = 0;

      while ( _lsp_templates[ i ].abbrev != NIL )
      {
         _lsp_assign_template( _lsp_templates[i].abbrev, _lsp_templates[i].template );
         i++;
      }
   }
}


/*
 ** _lsp_expand()
 *
 *  DESCRIPTION:
 *    Attempts to perform 'lsp' language template expansion.  This function
 *    is normally bound to the <SPACE> key.  When invoked, it reads the 
 *    previous word.  If the word is in the list of defined templates,
 *    the template is extracted and inserted.  Some characters in the 
 *    templates have special meanings. They are:
 *
 *       \n - new line
 *       &  - cursor position after template insertion
 *       @  - backspace
 *       \  - take next character literally
 */
BOOL DLL
_lsp_expand( void )
{
   LPVOID htEntry;
   LPSTR  _lsp_string;
   LPSTR  str;

   MarkSavePos();

   /* Check to see if we are currently positioned inside a 'lsp' comment
    */
   if (_lsp_in_comment( -1L ) != -1)
      goto _lsp_expand_incomplete;

   str = BufReadStr( 65535 );     /* read entire line */

   _lsp_string = StrTrim( str, " \t" );
   StrFree( str );

   if (!_lsp_string)
      goto _lsp_expand_incomplete;

   /* Search for a keyword match in the hash table.
    */
   htEntry = HashFindEntry( _lsp_hash_templates, _lsp_string, strlen( _lsp_string ) );
   StrFree( _lsp_string );

   if (htEntry)
   {
      /* A template exists for the current keyword, delete the keyword
       * and insert the template.
       */
      MarkDropPos( 1 );
      str = *(LPSTR*)htEntry;
      BufDelToEOL();

      _ext_expand_template( str );

   } 
   else
      goto _lsp_expand_incomplete;
   return TRUE;

_lsp_expand_incomplete:
   MarkRestorePos();

   return FALSE;
}

/*
 ** _lsp_language_colors_update()
 *
 *  This function does an intermediate update of the colors within
 *  a range of line numbers (firstline and lastline).  It must also
 *  determine if the first and last line specified are within a
 *  comment and act appropriately.
 */
long DLL
_lsp_language_colors_update( long firstline, long lastline )
{
   long startoff;
   long count = 0;

   if (firstline && lastline)
   {
      /* Begin coloring at the previous comment if
       * currently located in the middle of one.
       */
      _PosInit( 0 );
      if (firstline == 1)
         _PosPrevLine( 0L );
      else
         _PosNextLine( firstline - 1);
      startoff = _PosQOffset();

      /* Clear all attributes for the line range and
       * the reset them within _lsp_language_colors.
       */
      AttrSetColor( firstline, lastline, 1L, 0x7FFFFFFFL, 0 );
      _PosInit( startoff );
      _lsp_language_colors( lastline - firstline + 1 );
      count = lastline - firstline + 1;
   }
   return count;
}

/*
 ** _lsp_language_colors()
 *
 * If numlines == -1, start from the top of the buffer and encode all of them,
 * otherwise start at the current location of the number of lines specified.
 *
 * It is assumed that if numlines != -1, that _PosInit() has been called
 * and numlines is relative to the current position.
 */
void DLL
_lsp_language_colors( long numlines )
{
   long  startoff;
   long  startline;
   long  endoff;
   long  inc;
   WORD  ch;
   LPSTR buf;
   int   i = 0;
   int   comment_color = ColorComments( -1 );
   int   keyword_color = ColorKeywords( -1 );

   buf = (LPSTR)MemAlloc( MAX_BUF_LEN + 3 );

   /* Start at the top of the buffer or current position 
    */
   if (numlines == -1)
   {
      _PosInit( 0 );  
      numlines = 0x7FFFFFFF;
   }
   else
   {
      startoff  = _PosQOffset();
      startline = _PosQLine();
      endoff    = _lsp_in_comment( startoff );
      if (endoff == -1)
         _PosInit( startoff );
      else
      {
         _PosInit( endoff );
         numlines += (startline - _PosQLine());
      }
   }
   ch = _PosCurrentChar();
   startoff = _PosQOffset();


   while ((ch != EOF_CHAR) && (numlines > 0))
   {
      if (ch == '\n')
         numlines--;

      if (ch == '"' || ch == '\'' || ch == '{')
      {
         WORD quote = (ch == '{') ? '}' : ch;

         i = 0;
         startoff = _PosQOffset();

         do 
         {
            if (i < MAX_BUF_LEN)
               buf[i++] = (char)ch;
            ch = _PosNextChar();
         } while ((ch != EOF_CHAR) && (ch != '\n') && (ch != quote));
         buf[i++] = (char)ch;

         if ((ch == '\"') || (ch == '}'))
         {
            ch = _PosNextChar();
            goto _lsp_check_keyword;
         }
         ch = _PosNextChar();
         i = 0;
      }
      else if (isalpha( ch ) || (ch == '_'))
      {
         WORD lastch = _PosPrevChar();

         if (lastch != EOF_CHAR)
            _PosNextChar();

         i = 0;
         startoff = _PosQOffset();

         do
         {
            if (i < MAX_BUF_LEN)
               buf[i++] = (char)ch;
            ch = _PosNextChar();
         } while (isalpha( ch ) || (ch == '_') || (ch == '!') || (ch == '?'));

         if ((lastch != ' ' 
                  && lastch != '\t' 
                  && lastch != '\r' 
                  && lastch != '\n' 
                  && !ispunct( lastch ))
               || (ch == '_'))
            i = 0;
         continue;
      }
      else if (i)
      {
         if (isalpha( buf[0] ))
         {
_lsp_check_keyword:
            if (HashFindEntry( _lsp_hash_names, buf, i ))
            {
               endoff = _PosQOffset();
               _PosInit( startoff );
               _PosSetColor( keyword_color, endoff - startoff );
               _PosInit( endoff );
            }
         }
         i = 0;
      } 

      if (ch == ';')
      {
         startoff = _PosQOffset();

         inc = _PosNextLine( 1 );
         numlines--;
         endoff = _PosQOffset();
         _PosInit( startoff );
         _PosSetColor( comment_color, endoff - startoff - inc );
         _PosInit( endoff );

         ch = _PosCurrentChar();
         continue;
      }
      ch = _PosNextChar();
   }
   MemFree( buf );
}



void DLL
_lsp_add_keyword( LPSTR keyword )
{
   if (keyword && *keyword)
   {
      LPSTR XFAR *lpf = (LPSTR XFAR *)HashGetEntry( _lsp_hash_names, 
               keyword, strlen( keyword ) );
      if (lpf)
         *lpf = StrNew( keyword );
   }
}


/*
 ** _lsp_assign_template()
 *
 *  PARAMETERS:
 *    abbrev   - the keyword abreviation used to locate a template
 *    template - the encoded template string associated with 'abbrev'
 *
 *  DESCRIPTION:
 *    Provide a means of adding or modifying 'lsp' template strings
 *    without changing the code.
 */
void DLL
_lsp_assign_template( LPSTR abbrev, LPSTR template )
{
	_ext_assign_template( _lsp_hash_templates, abbrev, template );
}


/*
 ** _lsp_indent()
 *
 *  DESCRIPTION:
 *    Attempts to perform 'lsp' language smart indenting.  This function
 *    will behave differently based on the following conditions:
 *
 *    1. Within a comment - normal indentation
 *
 *    2. Not positioned at end of line - normal indentation
 *
 *    3. Current line prefixed by a name in lsp_INDENT_STR - perform smart
 *       indentation.  
 */
BOOL DLL
_lsp_indent( void )
{
   long   curline;
   long   col;
   LPSTR  _lsp_string;
   LPSTR  str;
   long   matchpos;

   MarkSavePos();

   curline = BufQCurrentLine();

   /* Check to see if we are currently positioned inside a 'lsp' comment
    */
   if (_lsp_in_comment( -1L ) != -1)
      goto _lsp_indent_incomplete;

   /* Not in a comment so indent according to the current 'lsp' construct.
    */
   if (!SrchFind( "^[ \t]*[a-zA-Z_]+\\c([ \t]|$)", SEARCH_REGEX , NIL )
         || (BufQCurrentLine() != curline) )
      goto _lsp_indent_incomplete;

   /* Read the keyword at the beginning of the current line.
    */
   col = BufQCurrentCol();
   SrchFind( "(^|[^a-zA-Z_])\\c[a-zA-Z_]", SEARCH_REGEX , NIL );
   str = BufReadStr( col - BufQCurrentCol() );
   _lsp_string = StrTrim( str, " \t" );
   StrFree( str );

   /* Set prefix and suffix to \377 to perform a quick lookup of the name
    */
   str = StrNew( "\377" );
   _lsp_string = StrAppend( _lsp_string, str );
   str = StrAppend( str, _lsp_string );
   StrFree( _lsp_string );
   
   matchpos = StrMatch( str, LSP_INDENT_STR, SEARCH_IGCASE | SEARCH_FORWARD, NIL );
   StrFree( str );

   if (!matchpos)
      /* Not a 'lsp' keyword, do normal indentation. */
      goto _lsp_indent_incomplete;
      
   MarkRestorePos();

   _ext_insert_indented_EOL();
   BufInsertChar( '\t' );
   return TRUE;

   _lsp_indent_incomplete:
      MarkRestorePos();

   return FALSE;
}



/*--  Private functions --*/


/*
 ** _lsp_in_comment()
 *
 *  Determines whether or not the current position in the buffer is
 *  located within a 'lsp' comment by search backward for a // or /* 
 *  comment declaration.
 *
 *  Returns -1 if not in a comment 
 *          >= 0 indicates the offset in the buffer where the comment begins
 */
static long
_lsp_in_comment( long offset )
{
   WORD  ch;

   /* Check to see if we are currently positioned inside a 'lsp' comment
    */
   _PosInit( offset );
   ch = _PosCurrentChar();

   do
   {
      if (ch == '\n')
         return -1;
      else if (ch == ';')
         return _PosQOffset();  /* currrent line is in a comment */
      ch = _PosPrevChar();
   } while (ch != EOF_CHAR);

   return -1L;
}

